#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import math

from theanolm.backend.probfunctions import *

class TestProbFunctions(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_interpolate_linear(self):
        self.assertAlmostEqual(
            interpolate_linear(numpy.log(0.2), numpy.log(0.3), numpy.float(0.25)),
            numpy.log(0.25 * 0.2 + 0.75 * 0.3))
        self.assertAlmostEqual(
            interpolate_linear(numpy.float('-inf'), numpy.log(0.3), numpy.float(0.01)),
            numpy.log(0.3 * 0.99))
        self.assertEqual(
            interpolate_linear(numpy.float('-inf'), numpy.float(-10), numpy.float(0)),
            numpy.float(-10))
        self.assertAlmostEqual(
            interpolate_linear(numpy.log(0.3), numpy.float('-inf'), numpy.float(0.99)),
            numpy.log(0.3 * 0.99))
        self.assertEqual(
            interpolate_linear(numpy.float(-10), numpy.float('-inf'), numpy.float(1)),
            numpy.float(-10))
        self.assertAlmostEqual(
            interpolate_linear(numpy.float(-1001), numpy.float(-1002), numpy.float(0.25)),
            numpy.float(-1001.64263),  # ln(0.25 * exp(-1001) + 0.75 * exp(-1002))
            places=4)

    def test_interpolate_loglinear(self):
        self.assertEqual(
            interpolate_loglinear(numpy.float(-1001), numpy.float(-1002),
                                  numpy.float(0.25), numpy.float(0.75)),
            numpy.float(-1001.75))
        self.assertEqual(
            interpolate_loglinear(numpy.float('-inf'), numpy.float(-1002),
                                  numpy.float(0.25), numpy.float(0.75)),
            numpy.float('-inf'))
        self.assertEqual(
            interpolate_loglinear(numpy.float('-inf'), numpy.float(-1002),
                                  numpy.float(0), numpy.float(1)),
            numpy.float(-1002))
        self.assertEqual(
            interpolate_loglinear(numpy.float(-1001), numpy.float('-inf'),
                                  numpy.float(0.25), numpy.float(0.75)),
            numpy.float('-inf'))
        self.assertEqual(
            interpolate_loglinear(numpy.float(-1001), numpy.float('-inf'),
                                  numpy.float(1), numpy.float(0)),
            numpy.float(-1001))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

import numpy
import theano
import theano.tensor as tensor
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams

from theanolm.backend import UniformDistribution, LogUniformDistribution
from theanolm.backend import MultinomialDistribution

class TestClassDistribution(unittest.TestCase):
    def setUp(self):
        theano.config.compute_test_value = 'warn'

        M1 = 2147483647
        M2 = 2147462579
        random_seed = [
            numpy.random.randint(0, M1),
            numpy.random.randint(0, M1),
            numpy.random.randint(1, M1),
            numpy.random.randint(0, M2),
            numpy.random.randint(0, M2),
            numpy.random.randint(1, M2)]
        self.random = RandomStreams(random_seed)

    def tearDown(self):
        pass

    def test_uniform_distribution_sample(self):
        distribution = UniformDistribution(self.random, 100)
        sample_tensor = distribution.sample(200, 400)
        f = theano.function([], [sample_tensor])

        sample = f()[0]
        self.assertEqual(sample.ndim, 2)
        self.assertEqual(sample.shape[0], 200)
        self.assertEqual(sample.shape[1], 400)
        self.assertEqual(sample.dtype, 'int64')
        self.assertTrue(numpy.all(numpy.greater_equal(sample, 0)))
        self.assertTrue(numpy.all(numpy.less(sample, 100)))

    def test_uniform_distribution_probs(self):
        distribution = UniformDistribution(self.random, 10)
        x_tensor = tensor.vector(dtype='int64')
        x_tensor.tag.test_value = numpy.array([1, 3, 5])
        probs = distribution.probs(x_tensor)
        self.assertEqual(probs, 0.1)

    def test_log_uniform_distribution_sample(self):
        distribution = LogUniformDistribution(self.random, 100)
        sample_tensor = distribution.sample(200, 400)
        f = theano.function([], [sample_tensor])

        sample = f()[0]
        self.assertEqual(sample.ndim, 2)
        self.assertEqual(sample.shape[0], 200)
        self.assertEqual(sample.shape[1], 400)
        self.assertEqual(sample.dtype, 'int64')
        self.assertTrue(numpy.all(numpy.greater_equal(sample, 0)))
        self.assertTrue(numpy.all(numpy.less(sample, 100)))

    def test_log_uniform_distribution_probs(self):
        distribution = LogUniformDistribution(self.random, 10)
        x_tensor = tensor.vector(dtype='int64')
        x_tensor.tag.test_value = numpy.array([1, 3, 5])
        probs_tensor = distribution.probs(x_tensor)
        f = theano.function([x_tensor], [probs_tensor])

        x = numpy.arange(10)
        probs = f(x)[0]
        self.assertAlmostEqual(probs.sum(), 1.0)
        self.assertGreater(probs[0], probs[1])
        self.assertGreater(probs[1], probs[2])
        self.assertGreater(probs[2], probs[3])
        self.assertGreater(probs[3], probs[4])
        self.assertGreater(probs[4], probs[5])
        self.assertGreater(probs[5], probs[6])
        self.assertGreater(probs[6], probs[7])
        self.assertGreater(probs[7], probs[8])
        self.assertGreater(probs[8], probs[9])
        self.assertAlmostEqual(probs[0], numpy.log(2) / numpy.log(11))
        self.assertAlmostEqual(probs[2:5].sum(), (numpy.log(6) - numpy.log(3)) / numpy.log(11))

    def test_multinomial_distribution_sample(self):
        probs = numpy.array([0.1, 0.01, 0.001, 0.0001, 0.00001, 0.000001, 0.0000001, 0.00000001, 0.000000001, 0.0000000001])
        probs /= probs.sum()
        probs_tensor = theano.shared(probs)
        distribution = MultinomialDistribution(self.random, probs_tensor)
        sample_tensor = distribution.sample(200, 3)
        f = theano.function([], [sample_tensor])

        sample = f()[0]
        self.assertEqual(sample.ndim, 2)
        self.assertEqual(sample.shape[0], 200)
        self.assertEqual(sample.shape[1], 3)
        self.assertEqual(sample.dtype, 'int64')
        self.assertTrue(numpy.all(numpy.greater_equal(sample, 0)))
        self.assertTrue(numpy.all(numpy.less(sample, 10)))

    def test_multinomial_distribution_probs(self):
        probs = numpy.array([0.01, 0.02, 0.03, 0.04, 0.1, 0.3, 0.5])
        probs_tensor = theano.shared(probs)
        distribution = MultinomialDistribution(self.random, probs_tensor)
        x_tensor = tensor.vector(dtype='int64')
        x_tensor.tag.test_value = numpy.array([1, 3, 5])
        probs_tensor = distribution.probs(x_tensor)
        f = theano.function([x_tensor], [probs_tensor])

        x = numpy.arange(7)
        probs = f(x)[0]
        self.assertAlmostEqual(probs[0], 0.01)
        self.assertAlmostEqual(probs[1], 0.02)
        self.assertAlmostEqual(probs[2], 0.03)
        self.assertAlmostEqual(probs[3], 0.04)
        self.assertAlmostEqual(probs[4], 0.1)
        self.assertAlmostEqual(probs[5], 0.3)
        self.assertAlmostEqual(probs[6], 0.5)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import numpy
import theano
import theano.tensor as tensor

from theanolm.backend import conv1d, conv2d
from theanolm.backend import l1_norm, sum_of_squares

class Test(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_conv1d(self):
        x_tensor = tensor.tensor3("x")
        filter_tensor = tensor.tensor3("filter")
        y_tensor = conv1d(x_tensor, filter_tensor)
        f = theano.function([x_tensor, filter_tensor], y_tensor)

        num_samples = 5
        num_elements = 10
        num_features = 2
        num_filters = 2
        x = numpy.arange(num_samples * num_elements * num_features)
        x = x.reshape([num_samples, num_elements, num_features])
        # Filter 1 is applied on feature 2, filter 2 is applied on feature 1.
        # Filter 1: [1]  Filter 2: [0]
        #           [0]            [1]
        #           [1]            [0]
        filters = numpy.array([[[0, 0],
                                [1, 0]],
                               [[0, 1],
                                [0, 0]],
                               [[0, 0],
                                [1, 0]]])

        y = f(x, filters)

        self.assertEqual(y.shape[0], num_samples)
        self.assertEqual(y.shape[1], num_elements - 2)
        self.assertEqual(y.shape[2], num_filters)

        for sample in range(num_samples):
            feature1_middle = (sample * 10 + numpy.arange(1, 9)) * 2
            feature2_left = (sample * 10 + numpy.arange(0, 8)) * 2 + 1
            feature2_right = (sample * 10 + numpy.arange(2, 10)) * 2 + 1
            self.assertSequenceEqual(list(y[sample, :, 0]),
                                     list(feature2_left + feature2_right))
            self.assertSequenceEqual(list(y[sample, :, 1]),
                                     list(feature1_middle))

    def test_conv2d(self):
        x_tensor = tensor.tensor4("x")
        filter_tensor = tensor.tensor4("filter")
        y_tensor = conv2d(x_tensor, filter_tensor, padding="same")
        f = theano.function([x_tensor, filter_tensor], y_tensor)

        num_samples = 5
        num_rows = 10
        num_columns = 10
        num_features = 2
        x = numpy.ones([num_samples, num_rows, num_columns, num_features])
        filters = numpy.zeros([3, 3, num_features, 7])

        y = f(x, filters)

        self.assertEqual(y.shape[0], num_samples)
        self.assertEqual(y.shape[1], num_rows)
        self.assertEqual(y.shape[2], num_columns)
        self.assertEqual(y.shape[3], 7)

    def test_l1_norm(self):
        a_tensor = tensor.matrix("a")
        b_tensor = tensor.matrix("b")
        y_tensor = l1_norm([a_tensor, b_tensor])
        f = theano.function([a_tensor, b_tensor], y_tensor)

        a = numpy.arange(9).reshape(3, 3)
        b = numpy.array([[-9, 10], [11, -12]])
        y = f(a, b)

        self.assertEqual(y, numpy.arange(13).sum())

    def test_sum_of_squares(self):
        a_tensor = tensor.matrix("a")
        b_tensor = tensor.matrix("b")
        y_tensor = sum_of_squares([a_tensor, b_tensor])
        f = theano.function([a_tensor, b_tensor], y_tensor)

        a = numpy.arange(9).reshape(3, 3)
        b = numpy.array([[9, -10], [-11, 12]])
        y = f(a, b)

        self.assertEqual(y, numpy.square(numpy.arange(13)).sum())

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import os

from theanolm.vocabulary import compute_word_counts, BigramStatistics
from theanolm import Vocabulary

class TestStatistics(unittest.TestCase):
    def setUp(self):
        script_path = os.path.dirname(os.path.realpath(__file__))
        sentences_path = os.path.join(script_path, 'sentences4.txt')
        self.sentences_file = open(sentences_path)

    def tearDown(self):
        self.sentences_file.close()

    def test_compute_word_counts(self):
        self.sentences_file.seek(0)
        word_counts = compute_word_counts([self.sentences_file])
        self.assertEqual(word_counts['a'], 13)
        self.assertEqual(word_counts['b'], 8)
        self.assertEqual(word_counts['c'], 8)
        self.assertEqual(word_counts['d'], 11)
        self.assertEqual(word_counts['e'], 15)
        self.assertEqual(word_counts['<s>'], 11)
        self.assertEqual(word_counts['</s>'], 11)

    def test_bigram_statistics(self):
        self.sentences_file.seek(0)
        word_counts = compute_word_counts([self.sentences_file])
        self.vocabulary = Vocabulary.from_word_counts(word_counts)
        self.sentences_file.seek(0)
        statistics = BigramStatistics([self.sentences_file], self.vocabulary)

        unigram_counts = statistics.unigram_counts
        vocabulary = self.vocabulary
        self.assertEqual(unigram_counts[vocabulary.word_to_id['a']], 13)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['b']], 8)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['c']], 8)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['d']], 11)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['e']], 15)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['<unk>']], 0)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['<s>']], 11)
        self.assertEqual(unigram_counts[vocabulary.word_to_id['</s>']], 11)

        bigram_counts = statistics.bigram_counts
        vocabulary = self.vocabulary
        a_id = vocabulary.word_to_id['a']
        b_id = vocabulary.word_to_id['b']
        self.assertEqual(bigram_counts[a_id,a_id], 3)
        self.assertEqual(bigram_counts[a_id,b_id], 2)
        self.assertEqual(bigram_counts[b_id,a_id], 1)
        self.assertEqual(bigram_counts[b_id,b_id], 0)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import os

import numpy
from numpy.testing import assert_almost_equal
import theano
from theano import tensor

from theanolm import Vocabulary
from theanolm.scoring import TextScorer

class DummyNetwork(object):
    """A dummy network for testing the text scorer that always outputs
    target_class_id / 100.0
    """

    def __init__(self, vocabulary):
        self.vocabulary = vocabulary
        self.input_word_ids = tensor.matrix('input_word_ids', dtype='int64')
        self.input_class_ids = tensor.matrix('input_class_ids', dtype='int64')
        self.target_class_ids = tensor.matrix('target_class_ids', dtype='int64')
        self.mask = tensor.matrix('mask', dtype='int64')
        self.oos_logprobs = theano.shared(
            numpy.log(vocabulary.get_oos_probs()),
            'network/oos_logprobs')
        self.is_training = tensor.scalar('is_training', dtype='int8')

    def target_probs(self):
        return self.target_class_ids.astype('float32') / 100.0

class TestTextScorer(unittest.TestCase):
    def setUp(self):
        script_path = os.path.dirname(os.path.realpath(__file__))
        vocabulary_path = os.path.join(script_path, 'vocabulary.txt')
        oos_words=['yksitoista', 'kaksitoista']
        with open(vocabulary_path) as vocabulary_file:
            self.vocabulary = Vocabulary.from_file(vocabulary_file, 'words',
                                                   oos_words=oos_words)
        word_counts = {'yksi': 1, 'kaksi': 2, 'kolme': 3, 'neljä': 4,
                       'viisi': 5, 'kuusi': 6, 'seitsemän': 7, 'kahdeksan': 8,
                       'yhdeksän': 9, 'kymmenen': 10, '<s>': 11, '</s>': 12,
                       '<unk>': 13, 'yksitoista': 3, 'kaksitoista': 7}
        self.vocabulary.compute_probs(word_counts)
        self.dummy_network = DummyNetwork(self.vocabulary)

    def tearDown(self):
        pass

    def test_score_batch(self):
        # Network predicts <unk> probability. Out-of-shortlist words are mapped
        # to <unk> class by .
        scorer = TextScorer(self.dummy_network, use_shortlist=False)
        word_ids = numpy.arange(15).reshape((3, 5)).T
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        mask = numpy.ones_like(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        assert_almost_equal(logprobs[0],
                            numpy.log(word_ids[1:,0].astype('float32') / 100.0))
        assert_almost_equal(logprobs[1],
                            numpy.log(word_ids[1:,1].astype('float32') / 100.0))
        self.assertAlmostEqual(logprobs[2][0], numpy.log(11.0 / 100.0), places=5)  # </s>
        self.assertAlmostEqual(logprobs[2][1], numpy.log(12.0 / 100.0), places=5)  # <unk>
        self.assertAlmostEqual(logprobs[2][2], numpy.log(12.0 / 100.0), places=5)
        self.assertAlmostEqual(logprobs[2][3], numpy.log(12.0 / 100.0), places=5)

        # Network predicts <unk> probability. This is distributed for
        # out-of-shortlist words according to word frequency.
        scorer = TextScorer(self.dummy_network, use_shortlist=True)
        word_ids = numpy.arange(15).reshape((3, 5)).T
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        mask = numpy.ones_like(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        assert_almost_equal(logprobs[0],
                            numpy.log(word_ids[1:,0].astype('float32') / 100.0))
        assert_almost_equal(logprobs[1],
                            numpy.log(word_ids[1:,1].astype('float32') / 100.0))
        self.assertAlmostEqual(logprobs[2][0], numpy.log(11.0 / 100.0), places=5)  # </s>
        self.assertIsNone(logprobs[2][1]) # <unk>
        self.assertAlmostEqual(logprobs[2][2], numpy.log(12.0 / 100.0 * 0.3), places=5)
        self.assertAlmostEqual(logprobs[2][3], numpy.log(12.0 / 100.0 * 0.7), places=5)

        # OOV and OOS words are replaced with None.
        scorer = TextScorer(self.dummy_network, use_shortlist=False,
                            exclude_unk=True)
        word_ids = numpy.arange(15).reshape((3, 5)).T
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        mask = numpy.ones_like(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        assert_almost_equal(logprobs[0],
                            numpy.log(word_ids[1:,0].astype('float32') / 100.0))
        assert_almost_equal(logprobs[1],
                            numpy.log(word_ids[1:,1].astype('float32') / 100.0))
        self.assertAlmostEqual(logprobs[2][0], numpy.log(11.0 / 100.0), places=5)  # </s>
        self.assertIsNone(logprobs[2][1]) # <unk>
        self.assertIsNone(logprobs[2][2])
        self.assertIsNone(logprobs[2][3])

    def test_score_sequence(self):
        # Network predicts <unk> probability.
        scorer = TextScorer(self.dummy_network, use_shortlist=False)
        word_ids = numpy.arange(15)
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        logprob = scorer.score_sequence(word_ids, class_ids, membership_probs)
        correct = word_ids[1:].astype('float32')
        correct /= 100.0
        correct[12] = 12.0 / 100.0
        correct[13] = 12.0 / 100.0
        correct = numpy.log(correct).sum()
        self.assertAlmostEqual(logprob, correct, places=4)

        # Network predicts <unk> probability. This is distributed for
        # out-of-shortlist words according to word frequency.
        scorer = TextScorer(self.dummy_network, use_shortlist=True)
        word_ids = numpy.arange(15)
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        logprob = scorer.score_sequence(word_ids, class_ids, membership_probs)
        correct = word_ids[1:].astype('float32')
        correct /= 100.0
        correct[11] = 1.0 # <unk> is ignored
        correct[12] = 12.0 / 100.0 * 0.3
        correct[13] = 12.0 / 100.0 * 0.7
        correct = numpy.log(correct).sum()
        self.assertAlmostEqual(logprob, correct, places=5)

        # OOV and OOS words are excluded from the resulting logprobs.
        scorer = TextScorer(self.dummy_network, use_shortlist=False,
                            exclude_unk=True)
        word_ids = numpy.arange(15)
        class_ids, _ = self.vocabulary.get_class_memberships(word_ids)
        membership_probs = numpy.ones_like(word_ids).astype('float32')
        logprob = scorer.score_sequence(word_ids, class_ids, membership_probs)
        correct = word_ids[1:12].astype('float32')
        correct /= 100.0
        correct = numpy.log(correct).sum()
        self.assertAlmostEqual(logprob, correct, places=5)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import os
import mmap

import numpy
from numpy.testing import assert_equal

from theanolm import Vocabulary
from theanolm.parsing import LinearBatchIterator, ScoringBatchIterator
from theanolm.parsing import ShufflingBatchIterator
from theanolm.parsing.functions import find_sentence_starts

class TestIterators(unittest.TestCase):
    def setUp(self):
        script_path = os.path.dirname(os.path.realpath(__file__))
        sentences1_path = os.path.join(script_path, 'sentences1.txt')
        sentences2_path = os.path.join(script_path, 'sentences2.txt')
        sentences3_path = os.path.join(script_path, 'sentences3.txt')
        vocabulary_path = os.path.join(script_path, 'vocabulary.txt')

        self.sentences1_file = open(sentences1_path)
        self.sentences2_file = open(sentences2_path)
        self.sentences3_file = open(sentences3_path)
        self.vocabulary_file = open(vocabulary_path)
        self.vocabulary = Vocabulary.from_file(self.vocabulary_file, 'words')
        self.vocabulary_file.seek(0)
        self.shortlist_vocabulary = \
            Vocabulary.from_file(self.vocabulary_file, 'words',
                                 oos_words=['yksitoista'])

    def tearDown(self):
        self.sentences1_file.close()
        self.sentences2_file.close()
        self.sentences3_file.close()
        self.vocabulary_file.close()

    def test_find_sentence_starts(self):
        sentences1_mmap = mmap.mmap(self.sentences1_file.fileno(),
                                    0,
                                    access=mmap.ACCESS_READ)
        sentence_starts = find_sentence_starts(sentences1_mmap)
        self.sentences1_file.seek(sentence_starts[0])
        self.assertEqual(self.sentences1_file.readline(), 'yksi kaksi\n')
        self.sentences1_file.seek(sentence_starts[1])
        self.assertEqual(self.sentences1_file.readline(), 'kolme neljä viisi\n')
        self.sentences1_file.seek(sentence_starts[2])
        self.assertEqual(self.sentences1_file.readline(), 'kuusi seitsemän kahdeksan\n')
        self.sentences1_file.seek(sentence_starts[3])
        self.assertEqual(self.sentences1_file.readline(), 'yhdeksän\n')
        self.sentences1_file.seek(sentence_starts[4])
        self.assertEqual(self.sentences1_file.readline(), 'kymmenen\n')
        self.sentences1_file.seek(0)

        sentences2_mmap = mmap.mmap(self.sentences2_file.fileno(),
                                    0,
                                    access=mmap.ACCESS_READ)
        sentence_starts = find_sentence_starts(sentences2_mmap)
        self.sentences2_file.seek(sentence_starts[0])
        self.assertEqual(self.sentences2_file.readline(), 'kymmenen yhdeksän\n')
        self.sentences2_file.seek(sentence_starts[1])
        self.assertEqual(self.sentences2_file.readline(), 'kahdeksan seitsemän kuusi\n')
        self.sentences2_file.seek(sentence_starts[2])
        self.assertEqual(self.sentences2_file.readline(), 'viisi\n')
        self.sentences2_file.seek(sentence_starts[3])
        self.assertEqual(self.sentences2_file.readline(), 'neljä\n')
        self.sentences2_file.seek(sentence_starts[4])
        self.assertEqual(self.sentences2_file.readline(), 'kolme kaksi yksi\n')
        self.sentences2_file.seek(0)

    def test_shuffling_batch_iterator(self):
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file],
                                          [],
                                          self.vocabulary,
                                          batch_size=2,
                                          max_sequence_length=5)

        sentences1 = []
        files1 = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            for sequence in range(2):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                sequence_class_ids = class_ids[sequence_mask != 0,sequence]
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                assert_equal(sequence_word_ids, sequence_class_ids)
                sentences1.append(' '.join(self.vocabulary.id_to_word[sequence_word_ids]))
                files1.extend(sequence_file_ids)
        self.assertEqual(files1.count(0), 20)
        self.assertEqual(files1.count(1), 20)
        sentences1_str = ' '.join(sentences1)
        sentences1_sorted_str = ' '.join(sorted(sentences1))
        self.assertEqual(sentences1_sorted_str,
                         '<s> kahdeksan seitsemän kuusi </s> '
                         '<s> kolme kaksi yksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> kymmenen </s> '
                         '<s> kymmenen yhdeksän </s> '
                         '<s> neljä </s> '
                         '<s> viisi </s> '
                         '<s> yhdeksän </s> '
                         '<s> yksi kaksi </s>')
        self.assertEqual(len(iterator), 5)

        sentences2 = []
        files2 = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            for sequence in range(2):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                sequence_class_ids = class_ids[sequence_mask != 0,sequence]
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                assert_equal(sequence_word_ids, sequence_class_ids)
                sentences2.append(' '.join(self.vocabulary.id_to_word[sequence_word_ids]))
                files2.extend(sequence_file_ids)
        self.assertCountEqual(sentences1, sentences2)
        self.assertCountEqual(files1, files2)
        self.assertTrue(sentences1 != sentences2)
        self.assertTrue(files1 != files2)

        # The sentences are wraped so that we get more sequences if we limit
        # the maximum length. Sequences shorter than two words will be ignored.
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file],
                                          [],
                                          self.vocabulary,
                                          batch_size=2,
                                          max_sequence_length=4)
        self.assertEqual(len(iterator), 5)
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file],
                                          [],
                                          self.vocabulary,
                                          batch_size=2,
                                          max_sequence_length=3)
        self.assertEqual(len(iterator), 7)

        # Sample 2 and 4 sentences (40 % and 80 %).
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file],
                                          [0.4, 0.8],
                                          self.vocabulary,
                                          batch_size=1,
                                          max_sequence_length=5)
        self.assertEqual(len(iterator), 2 + 4)

        # Make sure there are no duplicates.
        self.assertSetEqual(set(iterator._order),
                            set(numpy.unique(iterator._order)))
        self.assertEqual(numpy.count_nonzero(iterator._order <= 4), 2)
        self.assertEqual(numpy.count_nonzero(iterator._order >= 5), 4)

        # Use shortlist and don't map OOS words to <unk>.
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file,
                                           self.sentences3_file],
                                          [],
                                          self.shortlist_vocabulary,
                                          batch_size=2,
                                          map_oos_to_unk=False)
        word_counts = self._compute_word_counts(iterator)
        self._assert_oos_counts(word_counts)

        # Use shortlist and map OOS words to <unk>.
        iterator = ShufflingBatchIterator([self.sentences1_file,
                                           self.sentences2_file,
                                           self.sentences3_file],
                                          [],
                                          self.shortlist_vocabulary,
                                          batch_size=2,
                                          map_oos_to_unk=True)
        word_counts = self._compute_word_counts(iterator)
        self._assert_shortlist_counts(word_counts)

    def test_linear_batch_iterator(self):
        iterator = LinearBatchIterator(self.sentences1_file,
                                                self.vocabulary,
                                                batch_size=2,
                                                max_sequence_length=5)
        words = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            assert_equal(file_ids, 0)
            for sequence in range(mask.shape[1]):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                words.extend(self.vocabulary.id_to_word[sequence_word_ids])
        self.assertEqual(' '.join(words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s>')

        iterator = LinearBatchIterator([self.sentences1_file,
                                        self.sentences2_file],
                                       self.vocabulary,
                                       batch_size=2,
                                       max_sequence_length=5)
        words = []
        all_file_ids = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            for sequence in range(mask.shape[1]):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                words.extend(self.vocabulary.id_to_word[sequence_word_ids])
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                all_file_ids.extend(sequence_file_ids)
        self.assertEqual(' '.join(words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s> '
                         '<s> kymmenen yhdeksän </s> '
                         '<s> kahdeksan seitsemän kuusi </s> '
                         '<s> viisi </s> '
                         '<s> neljä </s> '
                         '<s> kolme kaksi yksi </s>')
        assert_equal(all_file_ids, [0, 0, 0, 0,
                                    0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0,
                                    0, 0, 0,
                                    0, 0, 0,
                                    1, 1, 1, 1,
                                    1, 1, 1, 1, 1,
                                    1, 1, 1,
                                    1, 1, 1,
                                    1, 1, 1, 1, 1])

        # Use shortlist and don't map OOS words to <unk>.
        iterator = LinearBatchIterator([self.sentences1_file,
                                        self.sentences2_file,
                                        self.sentences3_file],
                                       self.shortlist_vocabulary,
                                       batch_size=2,
                                       map_oos_to_unk=False)
        word_counts = self._compute_word_counts(iterator)
        self._assert_oos_counts(word_counts)

        # Use shortlist and map OOS words to <unk>.
        iterator = LinearBatchIterator([self.sentences1_file,
                                        self.sentences2_file,
                                        self.sentences3_file],
                                       self.shortlist_vocabulary,
                                       batch_size=2,
                                       map_oos_to_unk=True)
        word_counts = self._compute_word_counts(iterator)
        self._assert_shortlist_counts(word_counts)

    def test_scoring_batch_iterator(self):
        iterator = ScoringBatchIterator(self.sentences1_file,
                                        self.vocabulary,
                                        batch_size=2,
                                        max_sequence_length=5)
        all_words = []
        for word_ids, words, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            for sequence in range(mask.shape[1]):
                sequence_words = words[sequence]
                all_words.extend(sequence_words)
        self.assertEqual(' '.join(all_words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s>')

        iterator = ScoringBatchIterator([self.sentences1_file,
                                         self.sentences2_file],
                                        self.vocabulary,
                                        batch_size=2,
                                        max_sequence_length=5)
        all_words = []
        for word_ids, words, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            for sequence in range(mask.shape[1]):
                sequence_words = words[sequence]
                all_words.extend(sequence_words)
        self.assertEqual(' '.join(all_words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s> '
                         '<s> kymmenen yhdeksän </s> '
                         '<s> kahdeksan seitsemän kuusi </s> '
                         '<s> viisi </s> '
                         '<s> neljä </s> '
                         '<s> kolme kaksi yksi </s>')

        # Use shortlist and don't map OOS words to <unk>.
        iterator = ScoringBatchIterator([self.sentences1_file,
                                         self.sentences2_file,
                                         self.sentences3_file],
                                        self.shortlist_vocabulary,
                                        batch_size=2,
                                        map_oos_to_unk=False)
        word_counts = self._compute_word_counts(iterator)
        self._assert_oos_counts(word_counts)

        # Use shortlist and map OOS words to <unk>.
        iterator = ScoringBatchIterator([self.sentences1_file,
                                         self.sentences2_file,
                                         self.sentences3_file],
                                        self.shortlist_vocabulary,
                                        batch_size=2,
                                        map_oos_to_unk=True)
        word_counts = self._compute_word_counts(iterator)
        self._assert_shortlist_counts(word_counts)

    def _compute_word_counts(self, iterator):
        """Compute words counts using ``iterator``.
        """

        result = numpy.zeros(self.shortlist_vocabulary.num_words(),
                             dtype='int64')
        for word_ids, file_ids, mask in iterator:
            word_ids = word_ids[mask == 1]
            numpy.add.at(result, word_ids, 1)
        return result

    def _assert_oos_counts(self, word_counts):
        """When not mapping OOS words to ``<unk>``, the shortlist words appear
        three times each, and the OOS word and ``<unk>`` appear once. Assert
        that ``word_counts`` matches this.
        """

        for word in ['yksi', 'kaksi', 'kolme', 'neljä', 'viisi', 'kuusi',
                     'seitsemän', 'kahdeksan', 'yhdeksän', 'kymmenen']:
            word_id = self.shortlist_vocabulary.word_to_id[word]
            self.assertEqual(word_counts[word_id], 3)
        word_id = self.shortlist_vocabulary.word_to_id['yksitoista']
        self.assertEqual(word_counts[word_id], 1)
        word_id = self.shortlist_vocabulary.word_to_id['<unk>']
        self.assertEqual(word_counts[word_id], 1)
        word_id = self.shortlist_vocabulary.word_to_id['<s>']
        self.assertEqual(word_counts[word_id], 15)
        word_id = self.shortlist_vocabulary.word_to_id['</s>']
        self.assertEqual(word_counts[word_id], 15)

    def _assert_shortlist_counts(self, word_counts):
        """When mapping OOS words to ``<unk>``, the shortlist words appear three
        times each, the OOS word appears zero times, and ``<unk>`` appear twice.
        Assert that ``word_counts`` matches this.
        """

        for word in ['yksi', 'kaksi', 'kolme', 'neljä', 'viisi', 'kuusi',
                     'seitsemän', 'kahdeksan', 'yhdeksän', 'kymmenen']:
            word_id = self.shortlist_vocabulary.word_to_id[word]
            self.assertEqual(word_counts[word_id], 3)
        word_id = self.shortlist_vocabulary.word_to_id['yksitoista']
        self.assertEqual(word_counts[word_id], 0)
        word_id = self.shortlist_vocabulary.word_to_id['<unk>']
        self.assertEqual(word_counts[word_id], 2)
        word_id = self.shortlist_vocabulary.word_to_id['<s>']
        self.assertEqual(word_counts[word_id], 15)
        word_id = self.shortlist_vocabulary.word_to_id['</s>']
        self.assertEqual(word_counts[word_id], 15)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

from theanolm.vocabulary.wordclass import WordClass

class TestWordClass(unittest.TestCase):
    def setUp(self):
        pass

    def test_normalize_probs(self):
        word_class = WordClass(1, 10, 0.5)
        word_class.add(11, 1.0)
        word_class.add(12, 0.5)
        word_class.normalize_probs()
        self.assertEqual(word_class.get_prob(10), 0.25)
        self.assertEqual(word_class.get_prob(11), 0.5)
        self.assertEqual(word_class.get_prob(12), 0.25)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest

import numpy
from numpy.testing import assert_equal, assert_almost_equal

from theanolm.network.basiclayer import BasicLayer

class DummyParameters(object):
    def __init__(self):
        self._vars = dict()
        self._devs = dict()

    def __getitem__(self, path):
        return self._vars[path]

    def add(self, path, value, device=None):
        self._vars[path] = value
        self._devs[path] = device

    def get_device(self, path):
        return self._devs[path]

class DummyLayer(BasicLayer):
    def __init__(self, layer_options):
        super().__init__(layer_options, None)
        self._params = DummyParameters()

    def create_structure(self):
        pass

class TestBasicLayer(unittest.TestCase):
    def setUp(self):
        self.layer_options = dict()
        self.layer_options['name'] = 'layer_name'
        self.layer_options['input_layers'] = []
        self.layer_options['devices'] = ['dev0', 'dev1', 'dev2']

    def tearDown(self):
        pass

    def test_param_path(self):
        layer = DummyLayer(self.layer_options)
        self.assertEqual(layer._param_path('var1'), 'layers/layer_name/var1')
        self.assertEqual(layer._param_path('var2', 'dev1'), 'layers/layer_name/var2/dev1')

    def test_get_param(self):
        layer = DummyLayer(self.layer_options)
        layer._params.add('layers/layer_name/var1', 1)
        layer._params.add('layers/layer_name/var2', 2)
        layer._params.add('layers/layer_name/var3/dev1', 3)
        layer._params.add('layers/layer_name/var3/dev2', 4)
        self.assertEqual(layer._get_param('var1'), 1)
        self.assertEqual(layer._get_param('var2'), 2)
        self.assertEqual(layer._get_param('var3', 'dev1'), 3)
        self.assertEqual(layer._get_param('var3', 'dev2'), 4)

    def test_init_weight(self):
        layer = DummyLayer(self.layer_options)

        # standard normal distribution
        layer._init_weight('weight1', (100, 200), split_to_devices=False)
        weight = layer._get_param('weight1')
        self.assertListEqual(list(weight.shape), [100, 200])
        self.assertLess(numpy.min(weight), -1.0)
        self.assertGreater(numpy.max(weight), 1.0)

        # scaled
        layer._init_weight('weight2', (10, 20), 0.001, split_to_devices=False)
        weight = layer._get_param('weight2')
        self.assertGreater(numpy.min(weight), -1.0)
        self.assertLess(numpy.max(weight), 1.0)

        # orthogonal
        layer._init_weight('weight3', (100, 100), 0.001, split_to_devices=False)
        weight = layer._get_param('weight3')
        assert_almost_equal(numpy.dot(weight, weight.T), numpy.identity(100))

        # tiled
        layer._init_weight('weight4', (100, 200), count=3,
                           split_to_devices=False)
        weight = layer._get_param('weight4')
        self.assertListEqual(list(weight.shape), [100, 600])

        # split
        layer._init_weight('weight5', (100, 600), split_to_devices=True)
        weight = layer._get_param('weight5', 'dev0')
        self.assertListEqual(list(weight.shape), [100, 200])
        weight = layer._get_param('weight5', 'dev1')
        self.assertListEqual(list(weight.shape), [100, 200])
        weight = layer._get_param('weight5', 'dev2')
        self.assertListEqual(list(weight.shape), [100, 200])
        self.assertEqual(layer._params.get_device('layers/layer_name/weight5/dev0'), 'dev0')
        self.assertEqual(layer._params.get_device('layers/layer_name/weight5/dev1'), 'dev1')
        self.assertEqual(layer._params.get_device('layers/layer_name/weight5/dev2'), 'dev2')

        layer._init_weight('weight6', (100, 600), count=2, split_to_devices=True)
        weight = layer._get_param('weight6', 'dev0')
        self.assertListEqual(list(weight.shape), [100, 400])
        weight = layer._get_param('weight6', 'dev1')
        self.assertListEqual(list(weight.shape), [100, 400])
        weight = layer._get_param('weight6', 'dev2')
        self.assertListEqual(list(weight.shape), [100, 400])
        self.assertEqual(layer._params.get_device('layers/layer_name/weight6/dev0'), 'dev0')
        self.assertEqual(layer._params.get_device('layers/layer_name/weight6/dev1'), 'dev1')
        self.assertEqual(layer._params.get_device('layers/layer_name/weight6/dev2'), 'dev2')

    def test_init_bias(self):
        layer = DummyLayer(self.layer_options)

        # zeros
        layer._init_bias('bias1', (10, 20), split_to_devices=False)
        bias = layer._get_param('bias1')
        assert_equal(bias, numpy.zeros((10, 20)))

        # single value
        layer._init_bias('bias2', (10, 20), 5, split_to_devices=False)
        bias = layer._get_param('bias2')
        assert_equal(bias, numpy.ones((10, 20)) * 5)

        # array
        layer._init_bias('bias3', (10, 20),
                         numpy.arange(10 * 20).reshape(10, 20),
                         split_to_devices=False)
        bias = layer._get_param('bias3')
        assert_equal(bias, numpy.arange(10 * 20).reshape(10, 20))

        # tiled
        layer._init_bias('bias4', (3, 2), value=[5, 6, 7],
                         split_to_devices=False)
        bias = layer._get_param('bias4')
        value = numpy.array([[5, 5, 6, 6, 7, 7],
                             [5, 5, 6, 6, 7, 7],
                             [5, 5, 6, 6, 7, 7]])
        assert_almost_equal(bias, value)

        # split
        layer._init_bias('bias5', (2, 3), value=[5, 6, 7],
                         split_to_devices=True)
        bias = layer._get_param('bias5', 'dev0')
        value = numpy.ones((2, 3)) * 5
        assert_almost_equal(bias, value)
        bias = layer._get_param('bias5', 'dev1')
        value = numpy.ones((2, 3)) * 6
        assert_almost_equal(bias, value)
        bias = layer._get_param('bias5', 'dev2')
        value = numpy.ones((2, 3)) * 7
        assert_almost_equal(bias, value)
        self.assertEqual(layer._params.get_device('layers/layer_name/bias5/dev0'), 'dev0')
        self.assertEqual(layer._params.get_device('layers/layer_name/bias5/dev1'), 'dev1')
        self.assertEqual(layer._params.get_device('layers/layer_name/bias5/dev2'), 'dev2')

    def test_split_per_device(self):
        layer = DummyLayer(self.layer_options)
        ranges = layer._split_per_device(10)
        self.assertEqual(len(ranges), 3)
        self.assertTrue(len(ranges[0]) == 3 or len(ranges[0]) == 4)
        self.assertTrue(len(ranges[1]) == 3 or len(ranges[1]) == 4)
        self.assertTrue(len(ranges[2]) == 3 or len(ranges[2]) == 4)
        self.assertEqual(sum(len(x) for x in ranges), 10)
        self.assertEqual(ranges[0][0], 0)
        self.assertEqual(ranges[1][0], ranges[0][-1] + 1)
        self.assertEqual(ranges[2][0], ranges[1][-1] + 1)
        self.assertEqual(ranges[2][-1], 9)

if __name__ == '__main__':
    unittest.main()
